#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Torchdrug generation algorithm."""

import logging
import os
from typing import ClassVar, Dict, Optional, TypeVar

from ....training_pipelines.core import TrainingPipelineArguments
from ....training_pipelines.torchdrug.core import TorchDrugSavingArguments
from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import GAFGenerator, GCPNGenerator, Generator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=str)


class TorchDrugGenerator(GeneratorAlgorithm[S, T]):
    def __init__(
        self, configuration: AlgorithmConfiguration, target: Optional[T] = None
    ):
        """TorchDrug generation algorithm.

        Args:
            configuration: domain and application specification, defining types
                and validations.  Currently supported algorithm versions are:
                "zinc250k_v0", "qed_v0" and "plogp_v0".
            target: unused since it is not a conditional generator.

        Example:
            An example for using a generative algorithm from TorchDrug:

                configuration = TorchDrugGCPN(algorithm_version="qed_v0")
                algorithm = TorchDrugGenerator(configuration=configuration)
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        super().__init__(
            configuration=configuration,
            target=None,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample batches.

        Args:
            configuration: helps to set up the application.
            target: context or condition for the generation. Unused in the algorithm.

        Returns:
            callable generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.sample

    def validate_configuration(
        self, configuration: AlgorithmConfiguration
    ) -> AlgorithmConfiguration:
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(TorchDrugGenerator)
class TorchDrugGCPN(AlgorithmConfiguration[str, None]):
    """
    Interface for TorchDrug Graph-convolutional policy network (GCPN) algorithm.
    Currently supported algorithm versions are "zinc250k_v0", "qed_v0" and "plogp_v0".
    """

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "zinc250k_v0"

    def get_conditional_generator(self, resources_path: str) -> GCPNGenerator:
        """Instantiate the actual generator implementation.
        Args:
            resources_path: local path to model files.
        Returns:
            instance with :meth:`sample<gt4sd.algorithms.generation.torchdrug.implementation.GCPNGenerator.sample>` method for generation.
        """
        self.generator = GCPNGenerator(resources_path=resources_path)
        return self.generator

    @classmethod
    def get_filepath_mappings_for_training_pipeline_arguments(
        cls, training_pipeline_arguments: TrainingPipelineArguments
    ) -> Dict[str, str]:
        """Get filepath mappings for the given training pipeline arguments.
        Args:
            training_pipeline_arguments: training pipeline arguments.
        Returns:
            a mapping between artifacts' files and training pipeline's output files.
        """
        if isinstance(training_pipeline_arguments, TorchDrugSavingArguments):

            task_name = (
                f"task={training_pipeline_arguments.task}_"
                if training_pipeline_arguments.task
                else ""
            )
            data_name = "data=" + (
                training_pipeline_arguments.dataset_name
                + "_"
                + training_pipeline_arguments.file_path.split(os.sep)[-1].split(".")[0]
                if training_pipeline_arguments.dataset_name == "custom"
                else training_pipeline_arguments.dataset_name
            )

            epochs = training_pipeline_arguments.epochs
            return {
                "weights.pkl": os.path.join(
                    training_pipeline_arguments.model_path,
                    training_pipeline_arguments.training_name,
                    f"gcpn_data={data_name}_{task_name}epoch={epochs}.pkl",
                )
            }
        else:
            return super().get_filepath_mappings_for_training_pipeline_arguments(
                training_pipeline_arguments
            )


@ApplicationsRegistry.register_algorithm_application(TorchDrugGenerator)
class TorchDrugGraphAF(AlgorithmConfiguration[str, None]):
    """
    Interface for TorchDrug flow-based autoregressive graph algorithm (GraphAF).
    Currently supported algorithm versions are "zinc250k_v0", "qed_v0" and "plogp_v0".
    """

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "zinc250k_v0"

    def get_conditional_generator(self, resources_path: str) -> GAFGenerator:
        """Instantiate the actual generator implementation.
        Args:
            resources_path: local path to model files.
        Returns:
            instance with :meth:`samples<gt4sd.algorithms.generation.torchdrug.implementation.GAFGenerator.sample>` method for generation.
        """
        self.generator = GAFGenerator(resources_path=resources_path)
        return self.generator

    @classmethod
    def get_filepath_mappings_for_training_pipeline_arguments(
        cls, training_pipeline_arguments: TrainingPipelineArguments
    ) -> Dict[str, str]:
        """Get filepath mappings for the given training pipeline arguments.
        Args:
            training_pipeline_arguments: training pipeline arguments.
        Returns:
            a mapping between artifacts' files and training pipeline's output files.
        """
        if isinstance(training_pipeline_arguments, TorchDrugSavingArguments):

            task_name = (
                f"task={training_pipeline_arguments.task}_"
                if training_pipeline_arguments.task
                else ""
            )
            data_name = "data=" + (
                training_pipeline_arguments.dataset_name
                + "_"
                + training_pipeline_arguments.file_path.split(os.sep)[-1].split(".")[0]
                if training_pipeline_arguments.dataset_name == "custom"
                else training_pipeline_arguments.dataset_name
            )

            epochs = training_pipeline_arguments.epochs
            return {
                "weights.pkl": os.path.join(
                    training_pipeline_arguments.model_path,
                    training_pipeline_arguments.training_name,
                    f"graphaf_data={data_name}_{task_name}epoch={epochs}.pkl",
                )
            }
        else:
            return super().get_filepath_mappings_for_training_pipeline_arguments(
                training_pipeline_arguments
            )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Regression Transformer training implementation."""
import json
import logging
import os
import shutil
import tempfile
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from terminator.collators import TRAIN_COLLATORS
from terminator.tokenization import ExpressionBertTokenizer
from terminator.trainer import CustomTrainer, get_trainer_dict
from transformers import (
    CONFIG_MAPPING,
    MODEL_WITH_LM_HEAD_MAPPING,
    AutoConfig,
    AutoModelWithLMHead,
    DataCollatorForPermutationLanguageModeling,
    LineByLineTextDataset,
    set_seed,
)

from ..core import TrainingPipeline, TrainingPipelineArguments
from .utils import (
    get_hf_training_arg_object,
    get_train_config_dict,
    prepare_datasets_from_files,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

MODEL_CONFIG_CLASSES = list(MODEL_WITH_LM_HEAD_MAPPING.keys())  # type: ignore
MODEL_TYPES = tuple(conf.model_type for conf in MODEL_CONFIG_CLASSES)


class RegressionTransformerTrainingPipeline(TrainingPipeline):
    """RegressionTransformer training pipeline."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for training a Regression Transformer (RT) model.
           For details see:
            Born, J., & Manica, M. (2023). Regression Transformer enables concurrent sequence
            regression and generation for molecular language modelling.
            `Nature Machine Intelligence`, 5(4), 432-444.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.
        """
        try:
            params = {**training_args, **dataset_args, **model_args}
            # Setup logging
            logging.basicConfig(
                format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
                datefmt="%m/%d/%Y %H:%M:%S",
                level=logging.INFO,
            )
            if not params["do_train"]:
                logger.info("Nothing to do.")
                return

            training_name = params.get("training_name", "rt_training")

            logger.info(f"Model with name {training_name} starts.")
            self.setup_model(params)

            # Register training_dataset and eval_dataset
            self.train_dataset, self.test_dataset = self.setup_dataset(**params)
            logger.info(
                f"# samples: {len(self.train_dataset)}, {len(self.test_dataset)}."
            )

            # Model logging
            num_params = sum(
                p.numel() for p in self.model.parameters() if p.requires_grad
            )
            typ = type(self.model)
            logger.info(f"# parameters: {num_params}. Model: {typ}")
            if typ != "xlnet":
                logger.warning(f"Full functionality only with XLNet; not {typ}")

            # Setup training configuration
            self.model.resize_token_embeddings(len(self.tokenizer))
            if training_args["alternate_steps"] <= 0:
                # No alternation of training objectives means: Vanilla PLM training
                data_collator = DataCollatorForPermutationLanguageModeling(
                    tokenizer=self.tokenizer,
                    plm_probability=training_args["plm_probability"],
                    max_span_length=training_args["max_span_length"],
                )
                alternating_collator = None
            else:
                data_collator = TRAIN_COLLATORS["property"](
                    tokenizer=self.tokenizer,
                    property_tokens=self.properties,
                    num_tokens_to_mask=None,
                    mask_token_order=None,
                )
                alternating_collator = TRAIN_COLLATORS[training_args["cg_collator"]](
                    tokenizer=self.tokenizer,
                    property_tokens=self.properties,
                    plm_probability=training_args["plm_probability"],
                    max_span_length=training_args["max_span_length"],
                    do_sample=False,
                    entity_separator_token=training_args["entity_separator_token"],
                    mask_entity_separator=training_args["mask_entity_separator"],
                    entity_to_mask=training_args["entity_to_mask"],
                )

            # Initialize our Trainer
            train_config = get_train_config_dict(training_args, set(self.properties))
            os.makedirs(params["output_dir"], exist_ok=True)
            with open(
                os.path.join(params["output_dir"], "training_config.json"), "w"
            ) as f:
                json.dump(train_config, f, indent="\t")

            # Create the inference.json
            inference_dict = {
                "property_token": self.properties,
                "example": self.example_sample,
                "property_ranges": {
                    p.name: [p.minimum, p.maximum]
                    for p in self.property_objects.values()
                },
                "normalize": [False] * len(self.properties),  # True not supported atm
                "max_span_length": training_args["max_span_length"],
                "property_mask_length": {
                    p.name: p.mask_length for p in self.property_objects.values()
                },
            }
            with open(os.path.join(params["output_dir"], "inference.json"), "w") as f:
                json.dump(inference_dict, f, indent="\t")

            custom_trainer_params = get_trainer_dict(self.model_params)
            hf_train_object = get_hf_training_arg_object(training_args)

            trainer = CustomTrainer(
                model=self.model,
                args=hf_train_object,
                data_collator=data_collator,
                train_dataset=self.train_dataset,
                eval_dataset=self.test_dataset,
                tokenizer=self.tokenizer,
                alternating_collator=alternating_collator,
                train_config=train_config,
                **custom_trainer_params,
            )

            trainer.train(model_path=params["output_dir"])
            trainer.save_model()  # type: ignore

        except Exception:
            logger.exception(
                "Exception occurred while running RegressionTransformerTrainingPipeline."
            )

    def setup_model(self, params: Dict[str, Any]):
        """
        Error handling and training setup routine.

        Args:
            params: A dictionary with all parameters to launch training.

        Raises:
            ValueError: If flawed values are passed.
        """

        if params.get("test_data_path") is None and params["do_eval"]:
            raise ValueError(
                "Cannot do evaluation without an evaluation data file. Either supply a "
                "file to --test_data_path or remove the --do_eval argument."
            )
        if params["output_dir"] is None:
            raise ValueError(
                "You have to specify an output directory for the trained model"
            )
        if (
            os.path.exists(params["output_dir"])
            and os.listdir(params["output_dir"])
            and not params["overwrite_output_dir"]
        ):
            raise ValueError(
                f"Output directory ({params['output_dir']}) exists and is not empty. "
                "Use --overwrite_output_dir to overcome."
            )

        # Set seed
        set_seed(params["seed"])

        # Load model configuration.
        # Priorities:
        #   1) If provided, take it from the model path: `model_path`.
        #   2) If provided, take model configuration from `config_name`.
        #   3) Instantiate a fresh model.

        if params["model_path"] is None and params["model_type"] is None:
            raise ValueError(
                "Either pass pretrained model via `model_path` or specify"
                "which model to use via `model_typ`."
            )
        if params["model_path"]:
            config = AutoConfig.from_pretrained(
                params["model_path"],
                cache_dir=params["cache_dir"],
            )
            self.model_params = config.__dict__

            self.model = AutoModelWithLMHead.from_pretrained(
                params["model_path"],
                from_tf=False,
                config=config,
                cache_dir=params["cache_dir"],
            )
            logger.info(f"Model restored from {params['model_path']}")

        elif params["config_name"] is not None:
            with open(params["config_name"], "r") as f:
                self.model_params = json.load(f)
            config = AutoConfig.from_pretrained(
                params["config_name"],
                cache_dir=params["cache_dir"],
                mem_len=self.model_params.get("mem_len", 1024),
            )

        else:
            config = CONFIG_MAPPING[params["model_type"]]()
            self.model_params = config.__dict__
            logger.warning(
                f"Instantiating a new config instance: {params['model_type']}."
            )

        # Load tokenizer
        if params["model_path"]:
            # If model_path was provided we load tokenizer from there
            self.tokenizer = ExpressionBertTokenizer.from_pretrained(
                params["model_path"], cache_dir=params["cache_dir"]
            )
        elif params["tokenizer_name"]:
            self.tokenizer = ExpressionBertTokenizer.from_pretrained(
                params["tokenizer_name"], cache_dir=params["cache_dir"]
            )
        else:
            raise ValueError(
                f"No support for creating new tokenizer for: {params['model_type']}."
            )

        if not params["model_path"]:
            logger.info("Training new model from scratch")
            self.model = AutoModelWithLMHead.from_config(config)

    def setup_dataset(
        self,
        train_data_path: str,
        test_data_path: str,
        augment: int = 0,
        save_datasets: bool = False,
        *args,
        **kwargs,
    ):
        """
        Constructs the dataset objects.


        Args:
            train_data_path: Path to `.csv` file. Has to have a `text` column and
                at least one column of numerical properties.
            train_data_path: Path to `.csv` file. Has to have a `text` column and
                at least one column of numerical properties.
            augment: How many times each training sample is augmented.
            save_datasets: Whether to save the datasets to disk (will be stored in
                same location as `train_data_path` and `test_data_path`).

        Returns:
            A tuple of train and test dataset.
        """

        logger.info("Preparing/reading data...")

        tokenizer, properties, train_data, test_data = prepare_datasets_from_files(
            self.tokenizer, train_data_path, test_data_path, augment=augment
        )
        self.tokenizer, self.property_objects = tokenizer, properties
        self.properties = list(properties.keys())
        self.example_sample = train_data[0]

        train_dataset = self.create_dataset_from_list(
            train_data,
            save_path=train_data_path.replace(".csv", ".txt")
            if save_datasets
            else None,
        )
        test_dataset = self.create_dataset_from_list(
            test_data,
            save_path=test_data_path.replace(".csv", ".txt") if save_datasets else None,
        )
        logger.info("Finished data setup.")
        return train_dataset, test_dataset

    def create_dataset_from_list(
        self, data: List[str], save_path: Optional[str] = None
    ) -> LineByLineTextDataset:
        """
        Creates a LineByLineTextDataset from a List of strings.

        Args:
            data: List of strings with the samples.
            save_path: Path to save the dataset to. Defaults to None, meaning
                the dataset will not be saved.
        Returns:
            The dataset.
        """
        # Write files to temporary location and create data
        with tempfile.TemporaryDirectory() as temp:
            f_name = os.path.join(temp, "tmp_data.txt")

            # Write file
            with open(f_name, "w") as f:
                for line in data:
                    f.write(line + "\n")

            # Create dataset
            dataset = LineByLineTextDataset(
                file_path=f_name, tokenizer=self.tokenizer, block_size=2**64
            )
            if save_path:
                shutil.copyfile(f_name, save_path)
        return dataset


@dataclass
class RegressionTransformerModelArguments(TrainingPipelineArguments):
    """Arguments pertaining to model instantiation."""

    __name__ = "model_args"

    model_path: Optional[str] = field(
        default=None,
        metadata={"help": "Path where the pretrained model artifacts are stored."},
    )
    tokenizer_name: Optional[str] = field(
        default=None,
        metadata={
            "help": "Pretrained tokenizer name or path. If not provided, will be "
            "inferred from `model_path`. If `model_path` is not provided either you "
            "have to pass a tokenizer."
        },
    )
    config_name: Optional[str] = field(
        default=None,
        metadata={
            "help": "Pretrained config name or path. But `model_path` takes preference."
        },
    )
    model_type: Optional[str] = field(
        default="xlnet",
        metadata={
            "help": "If training from scratch, pass a model type from the list: "
            f"{', '.join(MODEL_TYPES)}. If `model_path` is also provided, `model_path` "
            "takes preference."
        },
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={
            "help": "Where do you want to store the pretrained models downloaded from s3"
        },
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of PaccMann^RL conditional generators."""

import json
import logging
import os
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
from paccmann_chemistry.models import StackGRUDecoder, StackGRUEncoder, TeacherVAE
from paccmann_chemistry.utils.search import SamplingSearch
from paccmann_omics.encoders import ENCODER_FACTORY
from pytoda.smiles.smiles_language import SMILESLanguage
from rdkit import Chem

from ....domains.materials import MoleculeFormat, validate_molecules
from ....domains.materials.protein_encoding import PrimarySequenceEncoder
from ....frameworks.torch import device_claim
from ....frameworks.torch.vae import reparameterize

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ConditionalGenerator(ABC):
    """Abstract interface for a conditional generator."""

    #: device where the inference is running.
    device: torch.device
    #: temperature for the sampling.
    temperature: float
    #: maximum length of the generated molecules.
    generated_length: int

    #: parameters for the SELFIES generator.
    selfies_conditional_generator_params: dict
    #: SELFIES generator.
    selfies_conditional_generator: TeacherVAE
    #: SMILES language instance.
    smiles_language: SMILESLanguage

    generator_latent_size: int
    encoder_latent_size: int

    def get_smiles_from_latent(self, latent: torch.Tensor) -> List[str]:
        """Take samples from the latent space.

        Args:
            latent: latent vector tensor.

        Returns:
            SMILES list and indexes for the valid ones.
        """
        if self.generator_latent_size == 2 * self.encoder_latent_size:
            latent = latent.repeat(1, 1, 2)

        # generate molecules as tokens list
        generated_molecules = self.selfies_conditional_generator.generate(
            latent,
            prime_input=torch.tensor(
                [self.smiles_language.start_index], device=self.device
            ).long(),
            end_token=torch.tensor(
                [self.smiles_language.stop_index], device=self.device
            ).long(),
            generate_len=self.generated_length,
            search=SamplingSearch(temperature=self.temperature),
        )

        molecules = [
            self.smiles_language.token_indexes_to_smiles(generated_molecule.tolist())
            for generated_molecule in iter(generated_molecules)
        ]

        # convert SELFIES to SMILES
        if "selfies" in self.smiles_language.name:
            molecules = [
                self.smiles_language.selfies_to_smiles(a_selfies)
                for a_selfies in molecules
            ]
        return molecules

    @staticmethod
    def validate_molecules(smiles) -> Tuple[List[Chem.rdchem.Mol], List[int]]:
        return validate_molecules(pattern_list=smiles, input_type=MoleculeFormat.smiles)

    @abstractmethod
    def get_latent(self, condition: Any) -> torch.Tensor:
        pass

    def generate_batch(self, condition: Any) -> List[str]:
        logger.info("embedding condition and getting reparametrized latent samples")
        latent = self.get_latent(condition)
        logger.info("starting generation of molecules")
        # generate the molecules
        return self.get_smiles_from_latent(latent)


class ProteinSequenceConditionalGenerator(ConditionalGenerator):
    """
    Protein conditional generator as implemented in https://doi.org/10.1088/2632-2153/abe808
    (originally https://arxiv.org/abs/2005.13285).
    It generates highly binding and low toxic ligands.

    Attributes:
        samples_per_protein: number of points sampled per protein.
            It has to be greater than 1.
        protein_embedding_encoder_params: parameter for the protein embedding encoder.
        protein_embedding_encoder: protein embedding encoder.
    """

    def __init__(
        self,
        resources_path: str,
        temperature: float = 1.4,
        generated_length: int = 100,
        samples_per_protein: int = 100,
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Initialize the generator.

        Args:
            resources_path: directory where to find models and parameters.
            temperature: temperature for the sampling. Defaults to 1.4.
            generated_length: maximum length of the generated molecules.
                Defaults to 100.
            samples_per_protein: number of points sampled per protein.
                It has to be greater than 1. Defaults to 10.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        # device
        self.device = device_claim(device)
        # setting sampling parameters
        self.temperature = temperature
        self.generated_length = generated_length
        self.samples_per_protein = samples_per_protein
        # instantiate protein embedding encoder
        with open(os.path.join(resources_path, "protein_embedding_params.json")) as fp:
            self.protein_embedding_encoder_params = json.load(fp)
        self.protein_embedding_encoder = ENCODER_FACTORY["dense"](
            self.protein_embedding_encoder_params
        ).to(self.device)
        self.protein_embedding_encoder.load(
            os.path.join(resources_path, "protein_embedding_encoder.pt"),
            map_location=self.device,
        )
        self.protein_embedding_encoder.eval()
        self.encoder_latent_size = self.protein_embedding_encoder.latent_size
        # instantiate selfies conditional generator
        with open(
            os.path.join(resources_path, "selfies_conditional_generator.json")
        ) as fp:
            self.selfies_conditional_generator_params = json.load(fp)
        self.selfies_conditional_generator = TeacherVAE(
            StackGRUEncoder(self.selfies_conditional_generator_params),
            StackGRUDecoder(self.selfies_conditional_generator_params),
        ).to(self.device)
        self.selfies_conditional_generator.load(
            os.path.join(resources_path, "selfies_conditional_generator.pt"),
            map_location=self.device,
        )
        self.selfies_conditional_generator.eval()
        self.generator_latent_size = (
            self.selfies_conditional_generator.decoder.latent_dim
        )
        # loading SMILES language for decoding and conversion of SELFIES to SMILES
        self.smiles_language = SMILESLanguage.load(
            os.path.join(resources_path, "selfies_language.pkl")
        )
        # protein embedding from primary sequence (via tape)
        self.primary_sequence_embedder = PrimarySequenceEncoder(
            model_type="transformer",
            from_pretrained="bert-base",
            model_config_file=None,
            tokenizer="iupac",
        ).to(self.device)

    def get_latent(self, protein: str) -> torch.Tensor:
        """
        Given a protein generate the latent representation.

        Args:
            protein: the protein used as context/condition.

        Returns:
            the latent representation for the given context. It contains
                self.samples_per_protein repeats.
        """
        # encode embedded sequence once, ignore the returned dummy ids
        embeddings, _ = self.primary_sequence_embedder.forward([[protein]])
        protein_mu, protein_logvar = self.protein_embedding_encoder(
            embeddings.to(self.device)
        )

        # now stack as batch to generate different samples
        proteins_mu = torch.cat([protein_mu] * self.samples_per_protein, dim=0)
        proteins_logvar = torch.cat([protein_logvar] * self.samples_per_protein, dim=0)
        # get latent representation
        return torch.unsqueeze(reparameterize(proteins_mu, proteins_logvar), 0)

    def generate_batch(self, protein: str) -> List[str]:
        return super().generate_batch(condition=protein)


class TranscriptomicConditionalGenerator(ConditionalGenerator):
    """
    Transcriptomic conditional generator as implemented in https://doi.org/10.1016/j.isci.2021.102269
    (originally https://doi.org/10.1007/978-3-030-45257-5_18, https://arxiv.org/abs/1909.05114).
    It generates highly effective small molecules against transcriptomic progiles.

    Attributes:
        samples_per_profile: number of points sampled per profile.
            It has to be greater than 1.
        transcriptomic_encoder_params: parameter for the protein embedding encoder.
        transcriptomic_encoder: protein embedding encoder.
    """

    def __init__(
        self,
        resources_path: str,
        temperature: float = 1.4,
        generated_length: int = 100,
        samples_per_profile: int = 100,
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Initialize the generator.

        Args:
            resources_path: directory where to find models and parameters.
            temperature: temperature for the sampling. Defaults to 1.4.
            generated_length: maximum length of the generated molecules.
                Defaults to 100.
            samples_per_profile: number of points sampled per protein.
                It has to be greater than 1. Defaults to 10.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        # device
        self.device = device_claim(device)
        # setting sampling parameters
        self.temperature = temperature
        self.generated_length = generated_length
        self.samples_per_profile = samples_per_profile
        with open(os.path.join(resources_path, "genes.txt")) as fp:
            self.genes = [gene.strip() for gene in fp if gene]
        # instantiate protein embedding encoder
        with open(os.path.join(resources_path, "transcriptomic_params.json")) as fp:
            self.transcriptomic_encoder_params = json.load(fp)
        self.transcriptomic_encoder = ENCODER_FACTORY["dense"](
            self.transcriptomic_encoder_params
        ).to(self.device)
        self.transcriptomic_encoder.load(
            os.path.join(resources_path, "transcriptomic_encoder.pt"),
            map_location=self.device,
        )
        self.transcriptomic_encoder.eval()
        self.encoder_latent_size = self.transcriptomic_encoder.latent_size
        # instantiate selfies conditional generator
        with open(
            os.path.join(resources_path, "selfies_conditional_generator.json")
        ) as fp:
            self.selfies_conditional_generator_params = json.load(fp)
        self.selfies_conditional_generator = TeacherVAE(
            StackGRUEncoder(self.selfies_conditional_generator_params),
            StackGRUDecoder(self.selfies_conditional_generator_params),
        ).to(self.device)
        self.selfies_conditional_generator.load(
            os.path.join(resources_path, "selfies_conditional_generator.pt"),
            map_location=self.device,
        )
        self.selfies_conditional_generator.eval()
        self.generator_latent_size = (
            self.selfies_conditional_generator.decoder.latent_dim
        )
        # loading SMILES language for decoding and conversion of SELFIES to SMILES
        self.smiles_language = SMILESLanguage.load(
            os.path.join(resources_path, "selfies_language.pkl")
        )

    def get_latent(self, profile: Union[np.ndarray, pd.Series, str]) -> torch.Tensor:
        """
        Given a profile generate the latent representation.

        Args:
            profile: the profile used as context/condition.

        Raises:
            ValueError: in case the profile has a size mismatch with the genes panel.

        Returns:
            the latent representation for the given context. It contains
                self.samples_per_profile repeats.
        """
        if isinstance(profile, pd.Series):
            # make sure genes are sorted
            profile = profile[self.genes].values
        elif isinstance(profile, str):
            logger.warning("profile passed as string, serializing it to a list")
            profile = np.array(json.loads(profile))
        if profile.size != len(self.genes):
            raise ValueError(
                f"provided profile size ({profile.size}) does not match required size {len(self.genes)}"
            )
        # encode embedded progiles
        transcriptomic_mu, transcriptomic_logvar = self.transcriptomic_encoder(
            torch.from_numpy(
                np.vstack([profile] * self.samples_per_profile),
            )
            .float()
            .to(self.device)
        )
        # get latent representation
        return torch.unsqueeze(
            reparameterize(transcriptomic_mu, transcriptomic_logvar), 0
        )

    def generate_batch(self, profile: Union[np.ndarray, pd.Series]) -> List[str]:
        return super().generate_batch(condition=profile)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import argparse
import configparser
from typing import Any, Dict, Optional

import sentencepiece as _sentencepiece
import torch as _torch
import tensorflow as _tensorflow
from pytorch_lightning import Trainer

from ..ml.models import ARCHITECTURE_FACTORY
from .utils import convert_string_to_class

# imports that have to be loaded before lightning to avoid segfaults
_sentencepiece
_tensorflow
_torch


def parse_arguments_from_config(conf_file: Optional[str] = None) -> argparse.Namespace:
    """Parse arguments from configuration file.

    Args:
        conf_file: configuration file. Defaults to None, a.k.a. us a default configuration
            in ./config/config.ini.

    Returns:
        the parsed arguments.
    """
    parser = argparse.ArgumentParser()

    # open config.ini file, either from parser or default file
    parser.add_argument(
        "--conf_file",
        type=str,
        help=("config file for the defaults value"),
        default="./config/config.ini",
    )

    # Read config file
    args, remaining_argv = parser.parse_known_args()
    config = configparser.ConfigParser()

    if conf_file:
        config.read(conf_file)
    else:
        config.read(args.conf_file)

    # classes that are not model name
    general_config_classes = ["general", "trainer", "default"]

    # adding a list of all model name into the args
    result: Dict[str, Any] = dict()
    result["model_list"] = [
        i for i in list(config.keys()) if i.lower() not in general_config_classes
    ]
    for key in [*config.keys()]:
        # go trough all models parameter, replace the parsed ones from the the config files ones
        if key.lower() not in general_config_classes:
            model_type = config[key]["type"]
            params_from_configfile = dict(config[key])
            model = ARCHITECTURE_FACTORY[model_type.lower()]
            parser = model.add_model_specific_args(parser, key)
            args, _ = parser.parse_known_args()
            args_dictionary = vars(args)
            params_from_configfile["name"] = key

            for i in params_from_configfile:
                params_from_configfile[i] = convert_string_to_class(
                    params_from_configfile[i]
                )

            params_from_configfile.update(
                {
                    k[: -len(key) - 1]: v
                    for k, v in args_dictionary.items()
                    if v is not None and k.endswith("_" + key)
                }
            )

            result[key] = params_from_configfile

        elif key.lower() == "trainer" or key.lower() == "general":
            params_from_configfile = dict(config[key])
            for i in params_from_configfile:
                params_from_configfile[i] = convert_string_to_class(
                    params_from_configfile[i]
                )
            result.update(params_from_configfile)

    # parser Pytorch Trainer arguments
    parser = Trainer.add_argparse_args(parser)

    # adding basename as the name of the run
    parser.add_argument("--basename", type=str)
    parser.add_argument("--batch_size", type=int)
    parser.add_argument("--num_workers", type=int)
    parser.add_argument("--lr", type=float)
    parser.add_argument("--validation_split", type=float, default=None)
    parser.add_argument("--validation_indices_file", type=str)
    args_dictionary = vars(parser.parse_args(remaining_argv))
    result.update({k: v for k, v in args_dictionary.items() if v is not None})
    result_namespace = argparse.Namespace(**result)

    return result_namespace



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Crystals crf training utilities."""

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from ...frameworks.crystals_rfc.rf_classifier import RFC
from ..core import TrainingPipeline, TrainingPipelineArguments

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class CrystalsRFCTrainingPipeline(TrainingPipeline):
    """Crystals RFC training pipelines for crystals."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for Crystals RFC models.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Raises:
            NotImplementedError: the generic trainer does not implement the pipeline.
        """

        rfc = RFC(crystal_sys=model_args["sym"])
        df = rfc.load_data(file_name=dataset_args["datapath"])
        train_x, test_x, train_y, test_y = rfc.split_data(
            df, test_size=dataset_args["test_size"]
        )
        train_x, test_x, train_y, test_y = rfc.normalize_data(
            train_x, test_x, train_y, test_y
        )

        rfc.train(train_x, train_y)

        rfc.save(training_args["output_path"])


@dataclass
class CrystalsRFCDataArguments(TrainingPipelineArguments):
    """Data arguments related to crystals RFC trainer."""

    __name__ = "dataset_args"

    datapath: str = field(
        metadata={
            "help": "Path to the dataset."
            "The dataset should follow the directory structure as described in https://github.com/dilangaem/SemiconAI."
        },
    )
    test_size: Optional[int] = field(
        default=None, metadata={"help": "Testing set percentage."}
    )


@dataclass
class CrystalsRFCModelArguments(TrainingPipelineArguments):
    """Model arguments related to crystals RFC trainer."""

    __name__ = "model_args"

    sym: str = field(
        default="all",
        metadata={
            "help": "Crystal systems to be used. 'all' for all the crystal systems. Other seven options are: 'monoclinic', 'triclinic', 'orthorhombic', 'trigonal', 'hexagonal', 'cubic', 'tetragonal'"
        },
    )


@dataclass
class CrystalsRFCTrainingArguments(TrainingPipelineArguments):
    """Training arguments related to crystals RFC trainer."""

    __name__ = "training_args"

    output_path: str = field(
        default=".",
        metadata={"help": "Path to the store the checkpoints."},
    )


@dataclass
class CrystalsRFCSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to crystals RFC trainer."""

    __name__ = "saving_args"



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Callable

from rdkit import Chem
from guacamol.utils.descriptors import bertz as _bertz
from guacamol.utils.descriptors import (
    logP,
    mol_weight,
    num_aromatic_rings,
    num_atoms,
    num_H_acceptors,
    num_H_donors,
    num_rings,
    num_rotatable_bonds,
)
from guacamol.utils.descriptors import qed as _qed
from guacamol.utils.descriptors import tpsa as _tpsa
from paccmann_generator.drug_evaluators import (
    ESOL,
    SAS,
    Lipinski,
    PenalizedLogP,
    SCScore,
)
from rdkit.Chem.rdMolDescriptors import CalcNumAtomStereoCenters, CalcNumHeterocycles
from rdkit.Chem.Scaffolds.MurckoScaffold import MurckoScaffoldSmiles

from ...domains.materials import SmallMolecule
from ..utils import to_mol, to_smiles

# Instantiate classes for faster inference
_sas = SAS()
_sccore = SCScore(score_scale=5, fp_len=1024, fp_rad=2)
_esol = ESOL()
_lipinski = Lipinski()
_penalized_logp = PenalizedLogP()


def plogp(mol: SmallMolecule) -> float:
    """Calculate the penalized logP of a molecule. This is the logP minus the number of
    rings with > 6 atoms minus the SAS.

    Gómez-Bombarelli, R., Wei, J. N., Duvenaud, D., Hernández-Lobato, J. M., Sánchez-Lengeling, B., Sheberla, D., ... & Aspuru-Guzik, A. (2018).
    Automatic chemical design using a data-driven continuous representation of molecules.
    ACS central science, 4(2), 268-276.

    NOTE: Check the initial arXiv for the plogp reference: https://arxiv.org/abs/1610.02415v1
    """
    return _penalized_logp(mol)


def lipinski(mol: SmallMolecule) -> int:
    """
    Calculate whether a molecule adheres to the Lipinski-rule-of-5.
    A crude approximation of druglikeness.

    Lipinski, C. A., Lombardo, F., Dominy, B. W., & Feeney, P. J. (1997).
    Experimental and computational approaches to estimate solubility and permeability in
    drug discovery and development settings.
    Advanced drug delivery reviews, 23(1-3), 3-25.
    """
    return int(_lipinski(mol)[0])


def esol(mol: SmallMolecule) -> float:
    """Estimate the water solubility of a molecule.

    Delaney, J. S. (2004).
    ESOL: estimating aqueous solubility directly from molecular structure.
    Journal of chemical information and computer sciences, 44(3), 1000-1005.

    """
    return _esol(mol)


def scscore(mol: SmallMolecule) -> float:
    """Calculate the synthetic complexity score (SCScore) of a molecule.

    Coley, C. W., Rogers, L., Green, W. H., & Jensen, K. F. (2018).
    SCScore: synthetic complexity learned from a reaction corpus.
    Journal of chemical information and modeling, 58(2), 252-261.

    """
    return _sccore(mol)


def sas(mol: SmallMolecule) -> float:
    """Calculate the synthetic accessibility score (SAS) for a molecule.

    Ertl, P., & Schuffenhauer, A. (2009).
    Estimation of synthetic accessibility score of drug-like molecules based on molecular
    complexity and fragment contributions.
    Journal of cheminformatics, 1(1), 1-11.
    """
    return _sas(mol)


def bertz(mol: SmallMolecule) -> float:
    """Calculate Bertz index of a molecule.

    Bertz, S. H. (1981).
    The first general index of molecular complexity.
    Journal of the American Chemical Society, 103(12), 3599-3601.
    """
    return _bertz(to_mol(mol))


def tpsa(mol: SmallMolecule) -> float:
    """
    Calculate the total polar surface area of a molecule.

    Ertl, P., Rohde, B., & Selzer, P. (2000).
    Fast calculation of molecular polar surface area as a sum of fragment-based
    contributions and its application to the prediction of drug transport properties.
    Journal of medicinal chemistry, 43(20), 3714-3717.
    """
    return _tpsa(to_mol(mol))


def logp(mol: SmallMolecule) -> float:
    """
    Calculates the partition coefficient of a molecule.

    Wildman, S. A., & Crippen, G. M. (1999).
    Prediction of physicochemical parameters by atomic contributions.
    Journal of chemical information and computer sciences, 39(5), 868-873.

    """
    return logP(to_mol(mol))


def qed(mol: SmallMolecule) -> float:
    """
    Calculate the quantitative estimate of drug-likeness (QED) of a molecule.

    Bickerton, G. R., Paolini, G. V., Besnard, J., Muresan, S., & Hopkins, A. L. (2012).
    Quantifying the chemical beauty of drugs.
    Nature chemistry, 4(2), 90-98.
    """
    return _qed(to_mol(mol))


def number_of_h_acceptors(mol: SmallMolecule) -> int:
    """Calculate number of H acceptors of a molecule."""
    return num_H_acceptors(to_mol(mol))


def number_of_atoms(mol: SmallMolecule) -> int:
    """Calculate number of atoms of a molecule."""
    return num_atoms(to_mol(mol))


def number_of_h_donors(mol: SmallMolecule) -> int:
    """Calculate number of H donors of a molecule."""
    return num_H_donors(to_mol(mol))


def number_of_aromatic_rings(mol: SmallMolecule) -> int:
    """Calculate number of aromatic rings of a molecule."""
    return num_aromatic_rings(to_mol(mol))


def number_of_rings(mol: SmallMolecule) -> int:
    """Calculate number of rings of a molecule."""
    return num_rings(to_mol(mol))


def number_of_rotatable_bonds(mol: SmallMolecule) -> int:
    """Calculate number of rotatable bonds of a molecule."""
    return num_rotatable_bonds(to_mol(mol))


def number_of_large_rings(mol: SmallMolecule) -> int:
    """Calculate the amount of large rings (> 6 atoms) of a molecule."""
    mol = to_mol(mol)
    ringinfo = mol.GetRingInfo()  # type: ignore
    return len([x for x in ringinfo.AtomRings() if len(x) > 6])


def molecular_weight(mol: SmallMolecule) -> float:
    """Calculate molecular weight of a molecule."""
    return mol_weight(to_mol(mol))


def is_scaffold(mol: SmallMolecule) -> int:
    """Whether a molecule is identical to its Murcko Scaffold."""
    mol = to_mol(mol)
    smi = Chem.MolToSmiles(mol, canonical=True)
    return int(smi == MurckoScaffoldSmiles(mol=mol))


def number_of_heterocycles(mol: SmallMolecule) -> int:
    """The amount of heterocycles of a molecule."""
    return CalcNumHeterocycles(to_mol(mol))


def number_of_stereocenters(mol: SmallMolecule) -> int:
    """The amount of stereo centers of a molecule."""
    return CalcNumAtomStereoCenters(to_mol(mol))


def similarity_to_seed(
    mol: SmallMolecule, similarity_fn: Callable[[SmallMolecule], float]
) -> float:
    """Calculate the similarity of a molecule to a seed molecule.

    Example:
        An example::

            from gt4sd.properties.molecules import similarity_to_seed, get_similarity_fn
            func = get_similarity_fn(target_mol='CCO', fp_key='FCFP4')
            similarity_to_seed(mol='CCC', similarity_fn=func)
    """
    return similarity_fn(to_smiles(mol))


def activity_against_target(
    mol: SmallMolecule, affinity_fn: Callable[[SmallMolecule], float]
) -> float:
    """Calculate the activity of a molecule against a target molecule.

    Example:
        An example::

            from gt4sd.properties.molecules import activity_against_target, get_activity_fn
            func = get_activity_fn(target_mol='DRD2')
            activity_against_target(mol='CCC', affinity_fn=func)
    """
    return affinity_fn(to_smiles(mol))



#
# MIT License
#
# Copyright (c) 2023 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Molformer training utilities."""

import logging
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Optional, Tuple, Union

import sentencepiece as _sentencepiece
import torch as _torch
import tensorflow as _tensorflow
import importlib_resources
from gt4sd_molformer.finetune.finetune_pubchem_light import (
    LightningModule as RegressionLightningModule,
)
from gt4sd_molformer.finetune.finetune_pubchem_light import (
    PropertyPredictionDataModule as RegressionDataModule,
)
from gt4sd_molformer.finetune.finetune_pubchem_light_classification import (
    LightningModule as ClassificationLightningModule,
)
from gt4sd_molformer.finetune.finetune_pubchem_light_classification import (
    PropertyPredictionDataModule as ClassificationDataModule,
)
from gt4sd_molformer.finetune.finetune_pubchem_light_classification_multitask import (
    MultitaskModel,
    PropertyPredictionDataModule,
)
from gt4sd_molformer.finetune.ft_tokenizer.ft_tokenizer import MolTranBertTokenizer
from gt4sd_molformer.training.train_pubchem_light import (
    LightningModule as PretrainingModule,
)
from gt4sd_molformer.training.train_pubchem_light import MoleculeModule
from gt4sd_trainer.hf_pl.pytorch_lightning_trainer import (
    PyTorchLightningTrainingPipeline,
)
from pytorch_lightning import LightningDataModule, LightningModule
from pytorch_lightning.utilities import seed

from ...core import TrainingPipelineArguments

# imports that have to be loaded before lightning to avoid segfaults
_sentencepiece
_tensorflow
_torch

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class MolformerTrainingPipeline(PyTorchLightningTrainingPipeline):
    """Molformer training pipelines for crystals."""

    def __init__(
        self,
        **kwargs,
    ):

        super().__init__(**kwargs)

        self.modules_getter = {
            "pretraining": self.get_pretraining_modules,
            "classification": self.get_classification_modules,
            "multitask_classification": self.get_multitask_classification_modules,
            "regression": self.get_regression_modules,
        }

    def get_data_and_model_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
        **kwargs,
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for training.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """

        if model_args["type"] not in self.modules_getter:
            raise ValueError(f"Training type {model_args['type']} is not supported.")

        # alignments with gt4sd-molformer
        model_args["run_id"] = datetime.now().strftime("%m%d%Y%H%M")
        dataset_args["n_embd"] = model_args["n_embd"]
        model_args["dataset_names"] = "valid test".split()  # type: ignore
        model_args["dataset_name"] = dataset_args["dataset_name"]
        model_args["measure_name"] = dataset_args["measure_name"]
        model_args["mode"] = model_args["pooling_mode"]
        del model_args["pooling_mode"]

        seed.seed_everything(model_args["seed"])  # type: ignore

        data_module, model_module = self.modules_getter[model_args["type"]](  # type: ignore
            model_args, dataset_args
        )

        return data_module, model_module

    def get_pretraining_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for pretraing.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """

        model_args["max_len"] = dataset_args["max_len"]

        train_config = {
            "batch_size": dataset_args["batch_size"],
            "num_workers": dataset_args["num_workers"],
            "pin_memory": True,
        }

        data_module = MoleculeModule(
            dataset_args["max_len"], dataset_args["data_path"], train_config
        )
        data_module.setup()

        model_module = PretrainingModule(model_args, data_module.get_vocab())

        return data_module, model_module

    def get_classification_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for pretraing.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """

        # return data_module, model_module

        bert_vocab_path = (
            importlib_resources.files("gt4sd_molformer") / "finetune/bert_vocab.txt"
        )

        tokenizer = MolTranBertTokenizer(bert_vocab_path)
        data_module = ClassificationDataModule(dataset_args, tokenizer)

        if model_args["pretrained_path"] is not None:
            model_module = ClassificationLightningModule(
                model_args, tokenizer
            ).load_from_checkpoint(
                model_args["pretrained_path"],
                strict=False,
                config=model_args,
                tokenizer=tokenizer,
                vocab=len(tokenizer.vocab),
            )
        else:
            model_module = ClassificationLightningModule(model_args, tokenizer)

        return data_module, model_module

    def get_multitask_classification_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for pretraing.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """

        model_args["num_tasks"] = len(dataset_args["measure_names"])  # type: ignore

        if model_args["num_tasks"] == 0:
            raise ValueError("Missing class names.")

        bert_vocab_path = (
            importlib_resources.files("gt4sd_molformer") / "finetune/bert_vocab.txt"
        )

        tokenizer = MolTranBertTokenizer(bert_vocab_path)

        data_module = PropertyPredictionDataModule(dataset_args, tokenizer)

        if model_args["pretrained_path"] is not None:
            model_module = MultitaskModel(model_args, tokenizer).load_from_checkpoint(
                model_args["pretrained_path"],
                strict=False,
                config=model_args,
                tokenizer=tokenizer,
                vocab=len(tokenizer.vocab),
            )
        else:
            model_module = MultitaskModel(model_args, tokenizer)

        return data_module, model_module

    def get_regression_modules(
        self,
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for pretraing.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """

        bert_vocab_path = (
            importlib_resources.files("gt4sd_molformer") / "finetune/bert_vocab.txt"
        )

        tokenizer = MolTranBertTokenizer(bert_vocab_path)

        if model_args["pretrained_path"] is not None:

            model_module = RegressionLightningModule(
                model_args, tokenizer
            ).load_from_checkpoint(
                model_args["pretrained_path"],
                strict=False,
                config=model_args,
                tokenizer=tokenizer,
                vocab=len(tokenizer.vocab),
            )

        else:
            model_module = RegressionLightningModule(model_args, tokenizer)

        data_module = RegressionDataModule(dataset_args, tokenizer)

        return data_module, model_module


@dataclass
class MolformerDataArguments(TrainingPipelineArguments):
    """Data arguments related to Molformer trainer."""

    __name__ = "dataset_args"

    batch_size: int = field(default=512, metadata={"help": "Batch size."})

    data_path: str = field(
        default="", metadata={"help": "Pretraining - path to the data file."}
    )

    max_len: int = field(default=100, metadata={"help": "Max of length of SMILES."})
    train_load: Optional[str] = field(
        default=None, metadata={"help": "Where to load the model."}
    )
    num_workers: Optional[int] = field(
        default=1, metadata={"help": "Number of workers."}
    )
    dataset_name: str = field(
        default="sol",
        metadata={
            "help": "Finetuning - Name of the dataset to be found in the data root directory."
        },
    )
    measure_name: str = field(
        default="measure",
        metadata={"help": "Finetuning - Measure name to be used as groundtruth."},
    )
    data_root: str = field(
        default="my_data_root",
        metadata={"help": "Finetuning - Data root for the dataset."},
    )
    train_dataset_length: Optional[int] = field(
        default=None, metadata={"help": "Finetuning - Length of training dataset."}
    )
    eval_dataset_length: Optional[int] = field(
        default=None, metadata={"help": "Finetuning - Length of evaluation dataset."}
    )
    aug: bool = field(default=False, metadata={"help": "aug."})
    measure_names: List[str] = field(
        default_factory=lambda: [],
        metadata={"help": "Class names for multitask classification."},
    )


@dataclass
class MolformerModelArguments(TrainingPipelineArguments):
    """Model arguments related to Molformer trainer."""

    __name__ = "model_args"

    type: str = field(
        default="classification",
        metadata={
            "help": "The training type, for example pretraining or classification."
        },
    )

    n_head: int = field(default=8, metadata={"help": "GPT number of heads."})
    n_layer: int = field(default=12, metadata={"help": "GPT number of layers."})
    q_dropout: float = field(default=0.5, metadata={"help": "Encoder layers dropout."})
    d_dropout: float = field(default=0.1, metadata={"help": "Decoder layers dropout."})
    n_embd: int = field(default=768, metadata={"help": "Latent vector dimensionality."})
    fc_h: int = field(
        default=512, metadata={"help": "Fully connected hidden dimensionality."}
    )
    dropout: float = field(
        default=0.1, metadata={"help": "Dropout used in finetuning."}
    )
    dims: List[int] = field(default_factory=lambda: [])
    num_classes: Optional[int] = field(
        default=None, metadata={"help": "Finetuning - Number of classes"}
    )

    restart_path: str = field(
        default="", metadata={"help": "path to  trainer file to continue training."}
    )

    lr_start: float = field(default=3 * 1e-4, metadata={"help": "Initial lr value."})

    lr_multiplier: int = field(default=1, metadata={"help": "lr weight multiplier."})

    seed: int = field(default=12345, metadata={"help": "Seed."})

    min_len: int = field(
        default=1, metadata={"help": "minimum length to be generated."}
    )

    root_dir: str = field(default=".", metadata={"help": "location of root dir."})

    num_feats: int = field(
        default=32, metadata={"help": "number of random features for FAVOR+."}
    )
    pooling_mode: str = field(
        default="cls", metadata={"help": "type of pooling to use."}
    )
    fold: int = field(default=0, metadata={"help": "number of folds for fine tuning."})
    pretrained_path: Optional[str] = field(
        default=None, metadata={"help": "Path to the base pretrained model."}
    )
    results_dir: str = field(
        default=".",
        metadata={"help": "Path to save evaluation results during training."},
    )
    debug: bool = field(default=False, metadata={"help": "Debug training"})


@dataclass
class MolformerTrainingArguments(TrainingPipelineArguments):
    """Training arguments related to Molformer trainer."""

    __name__ = "pl_trainer_args"

    accumulate_grad_batches: int = field(
        default=1,
        metadata={
            "help": "Accumulates grads every k batches or as set up in the dict."
        },
    )
    strategy: str = field(
        default="ddp",
        metadata={
            "help": "The accelerator backend to use (previously known as distributed_backend)."
        },
    )
    gpus: int = field(default=-1, metadata={"help": "number of gpus to use."})
    max_epochs: int = field(default=1, metadata={"help": "max number of epochs."})
    monitor: Optional[str] = field(
        default=None,
        metadata={"help": "Quantity to monitor in order to store a checkpoint."},
    )
    save_top_k: int = field(
        default=1,
        metadata={
            "help": "The best k models according to the quantity monitored will be saved."
        },
    )
    mode: str = field(
        default="min",
        metadata={"help": "Quantity to monitor in order to store a checkpoint."},
    )
    every_n_train_steps: Optional[int] = field(
        default=None,
        metadata={"help": "Number of training steps between checkpoints."},
    )
    every_n_epochs: Optional[int] = field(
        default=None,
        metadata={"help": "Number of epochs between checkpoints."},
    )
    save_last: Optional[bool] = field(
        default=None,
        metadata={
            "help": "When True, always saves the model at the end of the epoch to a file last.ckpt"
        },
    )
    save_dir: Optional[str] = field(
        default="logs", metadata={"help": "Save directory for logs and output."}
    )
    basename: Optional[str] = field(
        default="lightning_logs", metadata={"help": "Experiment name."}
    )
    val_check_interval: float = field(
        default=1.0, metadata={"help": " How often to check the validation set."}
    )
    gradient_clip_val: float = field(
        default=50, metadata={"help": "Gradient clipping value."}
    )
    resume_from_checkpoint: Optional[str] = field(
        default=None,
        metadata={"help": "Path/URL of the checkpoint from which training is resumed."},
    )


@dataclass
class MolformerSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to Molformer trainer."""

    __name__ = "saving_args"



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""GuacaMol Baselines tests."""

from typing import ClassVar, Type

import pytest

from gt4sd.algorithms.conditional_generation.guacamol import (
    GraphGAGenerator,
    GraphMCTSGenerator,
    GuacaMolGenerator,
    SMILESGAGenerator,
    SMILESLSTMHCGenerator,
    SMILESLSTMPPOGenerator,
)
from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.registry import ApplicationsRegistry
from gt4sd.tests.utils import GT4SDTestSettings

TARGET = {"isomer_scorer": {"target": 5.0, "target_smile": "NCCCCC"}}
algorithm_parameters = {
    "smiles_ga": {"random_start": True},
    "graph_ga": {"random_start": True},
    "graph_mcts": {
        "init_smiles": "CC",
        "population_size": 5,
        "generations": 5,
        "num_sims": 10,
        "max_children": 5,
        "max_atoms": 10,
    },
    "smiles_lstm_hc": {
        "random_start": True,
        "mols_to_sample": 10,
        "keep_top": 5,
        "max_len": 2,
        "optimize_batch_size": 3,
        "n_epochs": 2,
    },
    "smiles_lstm_ppo": {"num_epochs": 2, "episode_size": 10, "optimize_batch_size": 2},
}

test_settings = GT4SDTestSettings.get_instance()


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            SMILESGAGenerator,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
        ),
        (
            GraphGAGenerator,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
        ),
        (
            GraphMCTSGenerator,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
        ),
        (
            SMILESLSTMHCGenerator,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
        ),
        (
            SMILESLSTMPPOGenerator,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
        ),
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [
        (SMILESGAGenerator),
        (GraphGAGenerator),
        (GraphMCTSGenerator),
        (SMILESLSTMHCGenerator),
        (SMILESLSTMPPOGenerator),
    ],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [
        (SMILESGAGenerator),
        (GraphGAGenerator),
        (GraphMCTSGenerator),
        (SMILESLSTMHCGenerator),
        (SMILESLSTMPPOGenerator),
    ],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert "v0" in versions


@pytest.mark.parametrize(
    "config, algorithm, algorithm_parameters",
    [
        pytest.param(
            SMILESGAGenerator,
            GuacaMolGenerator,
            algorithm_parameters["smiles_ga"],
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="slow_runtime"),
        ),
        (GraphGAGenerator, GuacaMolGenerator, algorithm_parameters["graph_ga"]),
        (GraphMCTSGenerator, GuacaMolGenerator, algorithm_parameters["graph_mcts"]),
        pytest.param(
            SMILESLSTMHCGenerator,
            GuacaMolGenerator,
            algorithm_parameters["smiles_lstm_hc"],
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="slow_runtime"),
        ),
        pytest.param(
            SMILESLSTMPPOGenerator,
            GuacaMolGenerator,
            algorithm_parameters["smiles_lstm_ppo"],
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="slow_runtime"),
        ),
    ],
)
def test_generation_via_import(config, algorithm, algorithm_parameters):
    parameters = {
        "batch_size": 1,
    }
    for param, value in algorithm_parameters.items():
        parameters[param] = value
    config = config(**parameters)
    algorithm = algorithm(configuration=config, target=TARGET)
    items = list(algorithm.sample(1))
    assert len(items) == 1


@pytest.mark.parametrize(
    "algorithm_application, algorithm_type, domain, algorithm_name, algorithm_parameters",
    [
        pytest.param(
            SMILESGAGenerator.__name__,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
            algorithm_parameters["smiles_ga"],
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="slow_runtime"),
        ),
        (
            GraphGAGenerator.__name__,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
            algorithm_parameters["graph_ga"],
        ),
        (
            GraphMCTSGenerator.__name__,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
            algorithm_parameters["graph_mcts"],
        ),
        pytest.param(
            SMILESLSTMHCGenerator.__name__,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
            algorithm_parameters["smiles_lstm_hc"],
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="slow_runtime"),
        ),
        pytest.param(
            SMILESLSTMPPOGenerator.__name__,
            "conditional_generation",
            "materials",
            GuacaMolGenerator.__name__,
            algorithm_parameters["smiles_lstm_ppo"],
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="slow_runtime"),
        ),
    ],
)
def test_generation_via_registry(
    algorithm_type,
    domain,
    algorithm_name,
    algorithm_application,
    algorithm_parameters,
):
    parameters = {
        "target": TARGET,
        "algorithm_type": algorithm_type,
        "domain": domain,
        "algorithm_name": algorithm_name,
        "algorithm_application": algorithm_application,
        "batch_size": 1,
    }
    for param, value in algorithm_parameters.items():
        parameters[param] = value
    algorithm = ApplicationsRegistry.get_application_instance(**parameters)
    items = list(algorithm.sample(1))
    assert len(items) == 1



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""HuggingFace Diffusers generation algorithm. Code and models adapted from https://github.com/huggingface/diffusers."""

import logging
from dataclasses import field
from typing import Any, ClassVar, Dict, Optional, Set, TypeVar, Union

from ...core import (
    AlgorithmConfiguration,
    GeneratorAlgorithm,
    Untargeted,
    get_configuration_class_with_attributes,
)
from ...registry import ApplicationsRegistry
from .implementation import MODEL_TYPES, SCHEDULER_TYPES, Generator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=str)


class DiffusersGenerationAlgorithm(GeneratorAlgorithm[S, T]):
    def __init__(
        self, configuration: AlgorithmConfiguration, target: Optional[S] = None
    ):
        """Diffusers generation algorithm.

        Args:
            configuration: domain and application
                specification, defining types and validations.
            target: none for untargeted generation.

        Example:
            An example for using a generative algorithm from Diffusers::

                configuration = GeneratorConfiguration()
                algorithm = DiffusersGenerationAlgorithm(configuration=configuration)
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,
            target=None,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample batches.

        Args:
            configuration: helps to set up the application.
            target: context or condition for the generation.

        Returns:
            callable generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.sample

    def validate_configuration(
        self, configuration: AlgorithmConfiguration
    ) -> AlgorithmConfiguration:
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(DiffusersGenerationAlgorithm)
class DiffusersConfiguration(AlgorithmConfiguration[str, None]):
    """Basic configuration for a diffusion algorithm."""

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "vision"

    modality: str = field(
        default="image",
        metadata=dict(
            description="Modality.  Supported: 'image', 'text', 'audio', 'molecule'."
        ),
    )

    model_type: str = field(
        default="diffusion",
        metadata=dict(
            description=f"Type of the model. Supported: {', '.join(MODEL_TYPES.keys())}"
        ),
    )

    scheduler_type: str = field(
        default="discrete",
        metadata=dict(
            description=f"Type of the noise scheduler. Supported: {', '.join(SCHEDULER_TYPES.keys())}"
        ),
    )

    prompt: Union[str, Dict[str, Any]] = field(  # type: ignore
        default=None,
        metadata=dict(description="Prompt for conditional generation."),
    )

    def get_conditional_generator(self, resources_path: str, **kwargs) -> Generator:
        return Generator(
            resources_path=resources_path,
            model_type=self.model_type,
            model_name=self.algorithm_version,
            scheduler_type=self.scheduler_type,
            prompt=self.prompt,
        )


@ApplicationsRegistry.register_algorithm_application(DiffusersGenerationAlgorithm)
class DDPMGenerator(DiffusersConfiguration):
    """DDPM - Configuration to generate using unconditional denoising diffusion models."""

    algorithm_version: str = "google/ddpm-cifar10-32"
    model_type: str = "diffusion"
    scheduler_type: str = "ddpm"
    modality: str = "image"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://github.com/huggingface/diffusers"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(DiffusersGenerationAlgorithm)
class DDIMGenerator(DiffusersConfiguration):
    """DDIM - Configuration to generate using a denoising diffusion implicit model."""

    algorithm_version: str = "dboshardy/ddim-butterflies-128"
    model_type: str = "diffusion_implicit"
    scheduler_type: str = "ddim"
    modality: str = "image"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://github.com/huggingface/diffusers"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(DiffusersGenerationAlgorithm)
class LDMGenerator(DiffusersConfiguration):
    """Unconditional Latent Diffusion Model - Configuration to generate using a latent diffusion model."""

    algorithm_version: str = "CompVis/ldm-celebahq-256"
    model_type: str = "latent_diffusion"
    scheduler_type: str = "discrete"
    modality: str = "image"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://github.com/huggingface/diffusers"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(DiffusersGenerationAlgorithm)
class ScoreSdeGenerator(DiffusersConfiguration):
    """Score SDE Generative Model - Configuration to generate using a score-based diffusion generative model."""

    algorithm_version: str = "google/ncsnpp-celebahq-256"
    model_type: str = "score_sde"
    scheduler_type: str = "continuous"
    modality: str = "image"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://github.com/huggingface/diffusers"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(DiffusersGenerationAlgorithm)
class LDMTextToImageGenerator(DiffusersConfiguration):
    """Conditional Latent Diffusion Model - Configuration for conditional text2image generation using a latent diffusion model."""

    algorithm_version: str = "CompVis/ldm-text2im-large-256"
    model_type: str = "latent_diffusion_conditional"
    scheduler_type: str = "discrete"
    modality: str = "token2image"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://github.com/huggingface/diffusers"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(DiffusersGenerationAlgorithm)
class StableDiffusionGenerator(DiffusersConfiguration):
    """Stable Diffusion Model - Configuration for conditional text2image generation using a stable diffusion model.
    You have to provide authentication credentials to use this model.
    """

    algorithm_version: str = "CompVis/stable-diffusion-v1-4"
    model_type: str = "stable_diffusion"
    scheduler_type: str = "discrete"
    modality: str = "token2image"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://github.com/huggingface/diffusers"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(DiffusersGenerationAlgorithm)
class GeoDiffGenerator(DiffusersConfiguration):
    """GeoDiff Diffusion Model - Configuration for conditional 3D molecule structure generation given 2D information using a GeoDiff diffusion model."""

    algorithm_version: str = "fusing/gfn-molecule-gen-drugs"
    model_type: str = "geodiff"
    scheduler_type: str = "ddpm"
    modality: str = "molecule"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://github.com/huggingface/diffusers"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
            target description, returns None in case no target is used.
        """
        return {
            "title": "Prompt 2d representation for the molecule.",
            "description": "A dictionary containing all the information to build a molecule graph. Supported keys: ['atom_type', 'bond_edge_index', 'edge_index', 'edge_order', 'edge_type', 'idx', 'is_bond', 'num_nodes_per_graph', 'num_pos_ref', 'nx', 'pos', 'pos_ref', 'rdmol', 'smiles']",
            "type": "dictionary",
        }



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Model initialization module."""

from typing import Dict, Type

from .activation import ACTIVATION_FACTORY  # noqa: F401
from .base_model import GranularBaseModel, GranularEncoderDecoderModel
from .loss import LOSS_FACTORY  # noqa: F401
from .mlp_auto_encoder import MlpAutoEncoder
from .mlp_predictor import MlpPredictor
from .no_encoding import NoEncoding
from .vae_mlp import VaeMlp
from .vae_rnn import VaeRnn
from .vae_trans import VaeTrans

ARCHITECTURE_FACTORY: Dict[str, Type[GranularBaseModel]] = {
    "vae_rnn": VaeRnn,
    "vae_trans": VaeTrans,
    "mlp_predictor": MlpPredictor,
    "no_encoding": NoEncoding,
    "mlp_autoencoder": MlpAutoEncoder,
    "vae_mlp": VaeMlp,
}
AUTOENCODER_ARCHITECTURES = set(
    [
        model_type
        for model_type, model_class in ARCHITECTURE_FACTORY.items()
        if issubclass(model_class, GranularEncoderDecoderModel)
    ]
)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Recurrent Neural Networks with Proximal Policy Optimization algorithm implementation."""

from guacamol_baselines.smiles_lstm_ppo.goal_directed_generation import (
    PPODirectedGenerator,
)


class SMILESLSTMPPO:
    def __init__(
        self,
        model_path: str,
        num_epochs: int,
        episode_size: int,
        optimize_batch_size: int,
        entropy_weight: int,
        kl_div_weight: int,
        clip_param: float,
    ):
        """Initialize SMILESLSTMPPO.

        Args:
            model_path: path to load the model.
            num_epochs: number of epochs to sample.
            episode_size: number of molecules sampled by the policy at the start of a series of ppo updates.
            optimize_batch_size: batch size for the optimization.
            entropy_weight: used for calculating entropy loss.
            kl_div_weight: used for calculating Kullback-Leibler divergence loss.
            clip_param: used for determining how far the new policy is from the old one.
        """
        self.model_path = model_path
        self.num_epochs = num_epochs
        self.episode_size = episode_size
        self.optimize_batch_size = optimize_batch_size
        self.entropy_weight = entropy_weight
        self.kl_div_weight = kl_div_weight
        self.clip_param = clip_param

    def get_generator(self) -> PPODirectedGenerator:
        """Create an instance of the PPODirectedGenerator.

        Returns:
            an instance of PPODirectedGenerator.
        """
        optimiser = PPODirectedGenerator(
            pretrained_model_path=self.model_path,
            num_epochs=self.num_epochs,
            episode_size=self.episode_size,
            batch_size=self.optimize_batch_size,
            entropy_weight=self.entropy_weight,
            kl_div_weight=self.kl_div_weight,
            clip_param=self.clip_param,
        )
        return optimiser



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Tokenizers implementations."""

import collections
import logging
import os
from typing import Dict, Iterable, List, Type

import regex as re
import selfies as sf
from pytoda.smiles.processing import tokenize_selfies

SMI_REGEX_PATTERN = r"(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\|\/|:|~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"
BIG_SMI_REGEX_PATTERN = r"(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|\,|\{|\}|\[\]|=|#|-|\+|\\|\/|:|~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def selfies_alphabet() -> List[str]:
    """Legacy selfies 0.2.4 alphabet method.

    Adapted from: https://github.com/aspuru-guzik-group/selfies/blob/84122855ae76a928e1cb7d58796b8b47385a4359/selfies/selfies.py#L4.

    Returns:
        SELFIES list of tokens.
    """
    alphabet = [
        "[Branch1_1]",
        "[Branch1_2]",
        "[Branch1_3]",
        "[Ring1]",
        "[Branch2_1]",
        "[Branch2_2]",
        "[Branch2_3]",
        "[Ring2]",
        "[Branch3_1]",
        "[Branch3_2]",
        "[Branch3_3]",
        "[Ring3]",
        "[O]",
        "[=O]",
        "[N]",
        "[=N]",
        "[C]",
        "[=C]",
        "[#C]",
        "[S]",
        "[=S]",
        "[P]",
        "[F]",
        "[C@Hexpl]",
        "[C@@Hexpl]",
        "[C@expl]",
        "[C@@expl]",
        "[H]",
        "[NHexpl]",
    ]
    return alphabet


def load_vocab(vocab_file: str) -> Dict[str, int]:
    """Loads a vocabulary file into a dictionary.

    Args:
        vocab_file: vocabulary file.

    Returns:
        vocabulary mapping tokens to indices.
    """
    vocab = collections.OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip("\n")
        vocab[token] = index
    return vocab


class BasicTokenizer:
    """Basic tokenizer."""

    def __init__(
        self,
        pad_token: str = "<pad>",
        sos_token: str = "<sos>",
        eos_token: str = "</s>",
        unk_token: str = "<unk>",
    ) -> None:
        """Constructs a BasicSmilesTokenizer.

        Args:
            pad_token: padding token. Defaults to '<pad>'.
            sos_token: start of sequence token. Defaults to '<sos>'.
            eos_token: end of sequence token. Defaults to '</s>'.
            unk_token: unknown token. Defaults to '<unk>'.
        """
        self.pad_token = pad_token
        self.sos_token = sos_token
        self.eos_token = eos_token
        self.unk_token = unk_token

    def tokenize(self, text: str) -> List[str]:
        """Tokenize input text.

        Args:
            text: text to tokenize.

        Returns:
            list of tokens.
        """
        return list(text)

    def build_vocab(self, smiles: Iterable[str], vocab_file: str) -> List[str]:
        """Build and save a vocabulary given a SMILES list.

        Args:
            smiles: iterable of SMILES.
            vocab_file: path to a file where the vocabulary is saved.

        Returns:
            a list of all tokens in the vocabulary.
        """
        tokens = set([self.pad_token, self.sos_token, self.eos_token, self.unk_token])

        for smile in smiles:
            tokens_temp = self.tokenize(smile)

            for token in tokens_temp:
                tokens.add(token)

        tokens_list = sorted(list(tokens))

        with open(vocab_file, "w") as f:
            for item in tokens_list:
                f.write(f"{item}{os.linesep}")

        return tokens_list


class BasicSmilesTokenizer(BasicTokenizer):
    """Basic SMILES tokenizer."""

    def __init__(
        self,
        regex_pattern: str = SMI_REGEX_PATTERN,
        pad_token: str = "<pad>",
        sos_token: str = "<sos>",
        eos_token: str = "</s>",
        unk_token: str = "<unk>",
    ) -> None:
        """Constructs a BasicSmilesTokenizer.

        Args:
            regex_pattern: regex pattern. Defaults to SMI_REGEX_PATTERN.
            pad_token: padding token. Defaults to '<pad>'.
            sos_token: start of sequence token. Defaults to '<sos>'.
            eos_token: end of sequence token. Defaults to '</s>'.
            unk_token: unknown token. Defaults to '<unk>'.
        """
        self.regex_pattern = regex_pattern
        self.regex = re.compile(self.regex_pattern)
        super().__init__(
            pad_token=pad_token,
            sos_token=sos_token,
            eos_token=eos_token,
            unk_token=unk_token,
        )

    def tokenize(self, text: str) -> List[str]:
        """Tokenize input text.

        Args:
            text: text to tokenize.

        Returns:
            list of tokens.
        """
        return [token for token in self.regex.findall(text)]


class BasicSelfiesTokenizer(BasicTokenizer):
    """Basic SELFIES tokenizer."""

    def __init__(
        self,
        pad_token: str = "<pad>",
        sos_token: str = "<sos>",
        eos_token: str = "</s>",
        unk_token: str = "<unk>",
    ) -> None:
        """Constructs a BasicSelfiesTokenizer.

        Args:
            pad_token: padding token. Defaults to '<pad>'.
            sos_token: start of sequence token. Defaults to '<sos>'.
            eos_token: end of sequence token. Defaults to '</s>'.
            unk_token: unknown token. Defaults to '<unk>'.
        """
        self.pad_token = pad_token
        self.sos_token = sos_token
        self.eos_token = eos_token
        self.unk_token = unk_token

    def smiles_to_selfies(self, smiles: Iterable[str]) -> List[str]:
        """Convert a list of SMILES into SELFIES.

        Args:
            smiles: a list of SMILES.

        Returns:
            a list of SELFIES.
        """
        return [sf.encoder(a_smiles) for a_smiles in smiles]

    def tokenize(self, text: str) -> List[str]:
        """Tokenize input text.

        Args:
            text: text to tokenize.

        Returns:
            list of tokens.
        """
        return tokenize_selfies(sf.encoder(text))

    def build_vocab(self, smiles: Iterable[str], vocab_file: str) -> List[str]:
        """Build and save a vocabulary given a SMILES list.

        Args:
            smiles: iterable of SMILES.
            vocab_file: path to a file where the vocabulary is saved.

        Returns:
            a list of all tokens in the vocabulary.
        """
        selfies = self.smiles_to_selfies(smiles)
        tokens = set(
            [self.pad_token, self.sos_token, self.eos_token, self.unk_token, "[.]"]
            + selfies_alphabet()
        )
        for a_selfies in selfies:
            tokens = tokens | set(tokenize_selfies(a_selfies))

        tokens_list = sorted(list(tokens))

        with open(vocab_file, "w") as f:
            for item in tokens_list:
                f.write(f"{item}{os.linesep}")

        return tokens_list


class Tokenizer:
    """Tokenizer that can build a vocabulary on the fly."""

    def __init__(
        self,
        vocab_file: str,
        basic_tokenizer: BasicTokenizer = BasicTokenizer(),
        smiles: List[str] = [],
        pad_token: str = "<pad>",
        sos_token: str = "<sos>",
        eos_token: str = "</s>",
        unk_token: str = "<unk>",
    ) -> None:

        """Constructs a Tokenizer.

        Args:
            vocab_file: path to vocabulary file. If the file is not present, the provided SMILES list
                is used to generate one.
            basic_tokenizer: a basic tokenizer. Defaults to BasicTokenizer character tokenizer.
            smiles: list of smiles. Default to empty list, used only if the vocabulary file does not exist.
            pad_token: padding token. Defaults to '<pad>'.
            sos_token: start of sequence token. Defaults to '<sos>'.
            eos_token: end of sequence token. Defaults to '</s>'.
            unk_token: unknown token. Defaults to '<unk>'.
        """
        self.basic_tokenizer = basic_tokenizer
        self.pad_token = pad_token
        self.sos_token = sos_token
        self.eos_token = eos_token
        self.unk_token = unk_token

        # load or build vocab
        if os.path.isfile(vocab_file) and len(smiles) == 0:
            logger.info(f"load vocab from: {vocab_file}")
            self.vocab = load_vocab(vocab_file)
        else:
            logger.info("build tokenizer and vocabulary")
            self.basic_tokenizer.build_vocab(smiles, vocab_file)
            logger.info(f"saved vocabulary: {vocab_file}")
        self.vocab = load_vocab(vocab_file)
        self.vocab_ids = {token: index for token, index in self.vocab.items()}
        self.ids_to_tokens = collections.OrderedDict(
            [(ids, tok) for tok, ids in self.vocab.items()]
        )

        self.pad_token_id = self.vocab.get(pad_token, self.vocab[self.unk_token])
        self.sos_token_id = self.vocab.get(sos_token, self.vocab[self.unk_token])

    @property
    def vocab_size(self) -> int:
        """Size of the vocabulary.

        Returns:
            vocabulary file.
        """
        return len(self.vocab)

    @property
    def vocab_list(self) -> List[str]:
        """Return vocabulary tokens.

        Returns:
            all tokens from the vocabulary.
        """
        return list(self.vocab.keys())

    def tokenize(self, text: str) -> List[str]:
        """Tokenize a given text.

        Args:
            text: text to tokenize.

        Returns:
            list of tokens.
        """
        return [token for token in self.basic_tokenizer.tokenize(text)]

    def convert_tokens_to_ids(self, tokens: List[str]) -> List[int]:
        """Convert tokens to indices.

        Args:
            tokens: list of tokens.

        Returns:
            list of indices.
        """
        ids = []
        for token in tokens:
            ids.append(self.convert_token_to_id(token))
        return ids

    def convert_token_to_id(self, token: str) -> int:
        """Convert token to index.

        Args:
            token: a token.

        Returns:
            index corresponding to the input token. Unknown token index if the input
                token is not present in the vocabulary.
        """
        return self.vocab.get(token, self.vocab[self.unk_token])

    def convert_id_to_token(self, index: int) -> str:
        """Convert index to token.

        Args:
            index: an index.

        Returns:
            token corresponding to the input index. Unknown token if the input
                index is not found.
        """
        return self.ids_to_tokens.get(index, self.unk_token)

    def add_padding_tokens(
        self, token_ids: List[int], length: int, right: bool = True
    ) -> List[int]:
        """Add padding token indices to the provided token indices.

        Args:
            token_ids: token indices.
            length: length of the sequence.
            right: wheter the padding is performed on the right. Defaults to True, if False
                the padding happens on the left.

        Returns:
            the padded sequence.
        """
        padding = [self.pad_token_id] * (length - len(token_ids))
        if right:
            return token_ids + padding
        else:
            return padding + token_ids


class GenericTokenizer(Tokenizer):
    """Generic tokenizer that can build a vocabulary on the fly."""

    def __init__(
        self,
        vocab_file: str,
        smiles: List[str] = [],
        pad_token: str = "<pad>",
        sos_token: str = "<sos>",
        eos_token: str = "</s>",
        unk_token: str = "<unk>",
    ) -> None:
        """Constructs a GenericTokenizer.

        Args:
            vocab_file: path to vocabulary file. If the file is not present, the provided SMILES list
                is used to generate one.
            smiles: list of smiles. Default to empty list, used only if the vocabulary file does not exist.
            pad_token: padding token. Defaults to '<pad>'.
            sos_token: start of sequence token. Defaults to '<sos>'.
            eos_token: end of sequence token. Defaults to '</s>'.
            unk_token: unknown token. Defaults to '<unk>'.
        """
        super().__init__(
            vocab_file=vocab_file,
            basic_tokenizer=BasicTokenizer(
                pad_token=pad_token,
                sos_token=sos_token,
                eos_token=eos_token,
                unk_token=unk_token,
            ),
            smiles=smiles,
            pad_token=pad_token,
            sos_token=sos_token,
            eos_token=eos_token,
            unk_token=unk_token,
        )


class SmilesTokenizer(Tokenizer):
    """SMILES tokenizer that can build a vocabulary on the fly."""

    def __init__(
        self,
        vocab_file: str,
        smiles: List[str] = [],
        pad_token: str = "<pad>",
        sos_token: str = "<sos>",
        eos_token: str = "</s>",
        unk_token: str = "<unk>",
    ) -> None:
        """Constructs a SmilesTokenizer.

        Args:
            vocab_file: path to vocabulary file. If the file is not present, the provided SMILES list
                is used to generate one.
            smiles: list of smiles. Default to empty list, used only if the vocabulary file does not exist.
            pad_token: padding token. Defaults to '<pad>'.
            sos_token: start of sequence token. Defaults to '<sos>'.
            eos_token: end of sequence token. Defaults to '</s>'.
            unk_token: unknown token. Defaults to '<unk>'.
        """
        super().__init__(
            vocab_file=vocab_file,
            basic_tokenizer=BasicSmilesTokenizer(
                pad_token=pad_token,
                sos_token=sos_token,
                eos_token=eos_token,
                unk_token=unk_token,
            ),
            smiles=smiles,
            pad_token=pad_token,
            sos_token=sos_token,
            eos_token=eos_token,
            unk_token=unk_token,
        )


class BigSmilesTokenizer(Tokenizer):
    """Big-SMILES tokenizer that can build a vocabulary on the fly."""

    def __init__(
        self,
        vocab_file: str,
        smiles: List[str] = [],
        pad_token: str = "<pad>",
        sos_token: str = "<sos>",
        eos_token: str = "</s>",
        unk_token: str = "<unk>",
    ) -> None:
        """Constructs a BigSmilesTokenizer.

        Args:
            vocab_file: path to vocabulary file. If the file is not present, the provided Big-SMILES list
                is used to generate one.
            smiles: list of big smiles. Default to empty list, used only if the vocabulary file does not exist.
            pad_token: padding token. Defaults to '<pad>'.
            sos_token: start of sequence token. Defaults to '<sos>'.
            eos_token: end of sequence token. Defaults to '</s>'.
            unk_token: unknown token. Defaults to '<unk>'.
        """
        super().__init__(
            vocab_file=vocab_file,
            basic_tokenizer=BasicSmilesTokenizer(
                regex_pattern=BIG_SMI_REGEX_PATTERN,
                pad_token=pad_token,
                sos_token=sos_token,
                eos_token=eos_token,
                unk_token=unk_token,
            ),
            smiles=smiles,
            pad_token=pad_token,
            sos_token=sos_token,
            eos_token=eos_token,
            unk_token=unk_token,
        )


class SelfiesTokenizer(Tokenizer):
    """SELFIES tokenizer that can build a vocabulary on the fly."""

    def __init__(
        self,
        vocab_file: str,
        smiles: List[str] = [],
        pad_token: str = "<pad>",
        sos_token: str = "<sos>",
        eos_token: str = "</s>",
        unk_token: str = "<unk>",
    ) -> None:
        """Constructs a SelfiesTokenizer.

        Args:
            vocab_file: path to vocabulary file. If the file is not present, the provided SMILES list
                is used to generate one.
            smiles: list of smiles. Default to empty list, used only if the vocabulary file does not exist.
            pad_token: padding token. Defaults to '<pad>'.
            sos_token: start of sequence token. Defaults to '<sos>'.
            eos_token: end of sequence token. Defaults to '</s>'.
            unk_token: unknown token. Defaults to '<unk>'.
        """
        super().__init__(
            vocab_file=vocab_file,
            basic_tokenizer=BasicSelfiesTokenizer(
                pad_token=pad_token,
                sos_token=sos_token,
                eos_token=eos_token,
                unk_token=unk_token,
            ),
            smiles=smiles,
            pad_token=pad_token,
            sos_token=sos_token,
            eos_token=eos_token,
            unk_token=unk_token,
        )


TOKENIZER_FACTORY: Dict[str, Type[Tokenizer]] = {
    "generic": GenericTokenizer,
    "smiles": SmilesTokenizer,
    "big-smiles": BigSmilesTokenizer,
    "selfies": SelfiesTokenizer,
}
